Independent = {}

function Create(self)
	function Hash(Pos)
		return math.ceil(Pos.X/60) + math.floor(Pos.Y/60)*self.WidthFactor
	end
	
	self.LifeTimer = Timer()
	self.MoveTimer = Timer()
	self.MissileTimer = Timer()
	self.WatchdogTimer = Timer()
	self.missilePause = 250
	self.Gravity = SceneMan.GlobalAcc * TimerMan.DeltaTimeSecs	-- acceleration from gravity in pixels/frame/frame
	self.Factor = FrameMan.PPM * TimerMan.DeltaTimeSecs	-- converts m/s to pixels/frame
	self.WepRng = FrameMan.PlayerScreenWidth * 0.75
	self.TurretOffset = Vector(20, -7)
	self.ThrusterOffset = Vector(-16, 7)
	self.jetBurst = 7
	self.jetPower = 0.8
	self.acc_t = 0
	self.acc_ang = 0
	self.HeadAngle = 0
	self.WidthFactor = math.floor(SceneMan.SceneWidth / 60 + 0.5)
	self.resolution = 1
	
	if self.Team < 0 then
		self.Frame = 1
	else
		self.Frame = self.Team
	end
	
	if DebugIcon then
		DrawDebug = DebugIcon
	else
		DrawDebug = function(A, B) end
	end
	
	-- Find the head
	local MoObj
	for i = 1, MovableMan:GetMOIDCount()-1 do
		if i ~= thisID and MovableMan:GetRootMOID(i) == self.ID then
			MoObj = MovableMan:GetMOFromID(i)
			if MoObj.PresetName == "Independent Head" then
				self.Head = ToAttachable(MoObj)
				self.HeadAngle = self.Head.RotAngle
				break
			end
		end
	end
end

function Update(self)
	if self.Health < 1 then
		return
	end
	
	--[[ Debug
	if self.LZ then
		DrawDebug("Eye", self.LZ)
	end]]
	
	if not self.Head or not self.Head:IsAttached() then
		self:GibThis()
		return
	else
		-- Align head with target
		if self.HFlipped then
			self.Head.RotAngle = self.Head.RotAngle * 0.7 + (3.14 + self.HeadAngle) * 0.3
		else
			self.Head.RotAngle = self.Head.RotAngle * 0.7 + self.HeadAngle * 0.3
		end
	end
	
	local Ctrl = self:GetController()
	local TurretPos = self.Pos + self:RotateOffset(self.TurretOffset)
	local ThrusterPos = self.Pos + self:RotateOffset(self.ThrusterOffset)
	
	-- Update thrusters
	if self.Thrust then
		if MovableMan:IsParticle(self.Thrust) and self.Thrust.PresetName ~= "Entity" then
			self.Thrust.Pos = ThrusterPos
			self.Thrust.RotAngle = self.RotAngle +
				SceneMan:ShortestDistance(TurretPos + Vector(0,-15), ThrusterPos, false).AbsRadAngle
			
			if self.acc_t > 0 then
				self.Vel = self.Vel + Vector(self.jetPower,0):RadRotate(self.acc_ang)
				self.acc_t = self.acc_t - 1
			else
				self.Thrust:EnableEmission(false)
				if self.TargetVel and self.Vel.Magnitude > self.TargetVel.Magnitude then	-- We have probably left the ground
					self.WatchdogTimer:Reset()
					self.TargetVel = nil
				end
			end
		else
			self.Thrust = nil
		end
	end
	
	if self.Break then
		if MovableMan:IsParticle(self.Break) and self.Break.PresetName ~= "Entity" then
			self.Break.RotAngle = self.Vel.AbsRadAngle
			self.Break.Pos = self.Pos + self:RotateOffset(Vector(-8, 12))
			
			if self.Vel.Magnitude > 5 then
				self.Vel = Vector(self.Vel.X*0.85, self.Vel.Y*0.9)
			end
		else
			self.Break = nil
		end
	end
	
	-- Update muzzle flash
	if self.Flash then
		if MovableMan:IsParticle(self.Flash) and self.Flash.PresetName ~= "Entity" then
			self.Flash.Pos = TurretPos
		else
			self.Flash = nil
		end
	end
	
	-- Check if the target is still valid
	if self.Target and (not self.Target.ID or self.Target.ID > 254 or self.Target.PresetName == "Entity") then
		self.Target = nil
	end
	
	if self:IsPlayerControlled() then
		if Ctrl:IsState(Controller.BODY_JUMPSTART) then
			self.acc_t = 5
			if self.HFlipped then
				self.acc_ang = self.RotAngle+1.84
			else
				self.acc_ang = self.RotAngle+1.30
			end
			
			self.Vel = self.Vel + Vector(self.jetBurst/2,0):RadRotate(self.acc_ang)
			
			if self.Thrust then
				self.Thrust.ToDelete = true
			end
			
			-- Start jet pack
			self.Thrust = CreateAEmitter("Jump Thruster", "Independent.rte")
			self.Thrust.Pos = ThrusterPos
			self.Thrust.RotAngle = self.RotAngle +
				SceneMan:ShortestDistance(TurretPos + Vector(0,-15), ThrusterPos, false).AbsRadAngle
			MovableMan:AddParticle(self.Thrust)
		elseif Ctrl:IsState(Controller.BODY_JUMP) then
			self.acc_t = self.acc_t + 1
		end
		
		self.WatchdogTimer:Reset()
		return
	else
		if not self.WatchdogTimer:IsPastSimMS(20000) then
			Ctrl:SetState(Controller.MOVE_RIGHT, false)
			Ctrl:SetState(Controller.MOVE_LEFT, false)
			Ctrl:SetState(Controller.BODY_CROUCH, false)
		else
			--DrawDebug("Eye Green", self.Pos)	-- Debug
			
			if self.LZ then
				self:AddAISceneWaypoint(self.LZ)
				self.AIMode = Actor.AIMODE_GOTO
			elseif self.Targets and #self.Targets > 0 then
				local Act = self.Targets[math.random(#self.Targets)].Act
				if MovableMan:ValidMO(Act) and Act.PresetName ~= "Entity" then
					self:AddAISceneWaypoint(Act.Pos)
					self.AIMode = Actor.AIMODE_GOTO
				end
			else
				self.AIMode = Actor.AIMODE_BRAINHUNT
			end
			
			if Ctrl:IsState(Controller.BODY_JUMPSTART) then
				self.acc_t = 5
				if self.HFlipped then
					self.acc_ang = self.RotAngle+1.84
				else
					self.acc_ang = self.RotAngle+1.30
				end
			
				self.Vel = self.Vel*0.67 + Vector(self.jetBurst,0):RadRotate(self.acc_ang)
				
				if self.Thrust then
					self.Thrust.ToDelete = true
				end
				
				-- Start jet pack
				self.Thrust = CreateAEmitter("Jump Thruster", "Independent.rte")
				self.Thrust.Pos = ThrusterPos
				self.Thrust.RotAngle = self.RotAngle +
					SceneMan:ShortestDistance(TurretPos + Vector(0,-15), ThrusterPos, false).AbsRadAngle
				MovableMan:AddParticle(self.Thrust)
			end
		end
	end
	
	-- Stabilize
	self.RotAngle = self.RotAngle * 0.95
	if math.abs(self.RotAngle) < 0.2 or  math.abs(self.AngularVel) > 5 then
		self.AngularVel = self.AngularVel * 0.9
	end
	
	-- Search for targets and evaluate the threat level
	if (not self.Target and self.LifeTimer:IsPastSimMS(250)) or self.LifeTimer:IsPastSimMS(750) then
		self.LifeTimer:Reset()
		
		-- Dislodge
		if self.Vel.Magnitude < 0.1 then
			self:ForceDeepCheck(true)
		end
		
		if self.Health < 99 then
			self.Health = self.Health + RangeRand(0.4, 1.4)
		end

		local Act, Weapon, Dist, range, score, root_id
		self.Targets = {}
		self.Occupied = {}
		for i = 1, MovableMan:GetMOIDCount()-1 do	-- Identify enemy weapons by iterating over the MOs instead of the actors
			root_id = MovableMan:GetRootMOID(i)
			if root_id ~= self.ID then	-- Don't target yourself
				if root_id ~= i then	-- A held weapon have the root id of the actor holding it
					Act = MovableMan:GetMOFromID(root_id)
					if MovableMan:IsActor(Act) then
						Act = ToActor(Act)
						if Act.Team ~= self.Team and Act.PresetName ~= "Entity" then
							Dist = SceneMan:ShortestDistance(TurretPos, Act.Pos, false)
							range = Dist.Magnitude
							score = 1 - range / SceneMan.SceneDim.Magnitude + Act.GoldValue/5000 + Act.Health/1000
							
							-- Bias towards targeting the AI controlled actors
							if Act:IsPlayerControlled() then
								score = score - 0.05
							end
							
							if range < self.WepRng then
								score = score + 0.2	-- Within our range
								if SceneMan:CastFindMORay(TurretPos, Dist, Act.ID, Vector(), self.ID, false, 7) then
									score = score + 0.5	-- LOS
									if Act:GetController():IsState(Controller.AIM_SHARP) then
										Dist = SceneMan:ShortestDistance(self.Pos, Act.ViewPoint, false)
										score = score + 0.1
										if Dist.Magnitude < 60 then	-- This enemy is aiming at us
											score = score + 0.4
										end
									end
								end
							end
							
							if Act.ClassName == "ADoor" then
								score = score * 0.5
								Act = MovableMan:GetMOFromID(i)	-- Attack the actual door instead of the motor
							else
								Weapon = MovableMan:GetMOFromID(i)
								if Weapon.ClassName == "HDFirearm" then
									local tmp = math.min(0.01 + Weapon.GoldValue/1000 + Weapon.Mass/100, 0.15)
									if range < ToHeldDevice(Weapon).SharpLength then
										score = score + tmp * 2
									else
										score = score + tmp
									end
								end
								
								if self.InvTarget and self.InvTarget.ID == Act.ID then
									score = score / 3
									--DrawDebug("Eye Green Long", Act.Pos) --Debug
								end
							end
							
							table.insert(self.Targets, {Act=Act, score=score, dist=range})
						end
					end
				else
					Act = MovableMan:GetMOFromID(root_id)
					if MovableMan:IsActor(Act) and Act.PresetName ~= "Entity" and Act.ClassName ~= "ADoor" then
						Act = ToActor(Act)
						if Act.Team == self.Team then
							score = -1
						else
							range = SceneMan:ShortestDistance(TurretPos, Act.Pos, false).Magnitude
							score = (1 - range / SceneMan.SceneDim.Magnitude + Act.GoldValue/5000 + Act.Health/1000)*0.9
							table.insert(self.Targets, {Act=Act, score=score, dist=range})
						end
						
						-- Transform the position in to a spatial hash
						if Act.Vel.Magnitude < 3 then
							local h = Hash(Act.Pos)
							for i = -1, 1 do
								self.Occupied[h+i] = score
							end
						end
					end
				end
			end
		end
		
		if #self.Targets > 0 then
			self.targetIndex = math.min(math.floor(RangeRand(1,3)+0.5), #self.Targets)
			table.sort(self.Targets, function(a,b) return a.score > b.score end)
			self.InvTarget = nil
			self.Target = self.Targets[1].Act
		else
			self.Targets = nil
		end
	elseif self.Targets and self.MissileTimer:IsPastSimMS(self.missilePause) and self.Vel.Magnitude < 1 then
		if self.AimRocket then	-- Look for a missile trajectory to the target
			if not MovableMan:ValidMO(self.AimRocket.Act) or self.AimRocket.Act.PresetName == "Entity" then
				self.AimRocket = nil
			else
				local running, Data = coroutine.resume(self.AimRocket.Co,
					self,
					self.AimRocket.Act.Pos
				)
				
				if Data and Data.ang then
					self.HeadAngle = Data.ang
					if Data.str + Data.dist < self.AimRocket.minStr + self.AimRocket.dist and
						SceneMan:CastObstacleRay(self.Pos + self:RotateOffset(Vector(7,-20)),
							Vector(50,0):RadRotate(Data.ang), Vector(), Vector(), self.ID, -1, 3) == -1
					then
						self.AimRocket.minStr = Data.str
						self.AimRocket.dist = Data.dist
						self.AimRocket.ang = Data.ang
					end
				end
				
				if not running then	-- the coroutine return false when done
					self.MissileTimer:Reset()
					
					if self.AimRocket.ang and self.AimRocket.minStr < 70  then -- Launch
						local Missile
						if self.AimRocket.Act.Vel.Magnitude < 3.5 then
							if self.AimRocket.dist < 60 then
								Missile = CreateAEmitter("Rocket Launch", "Independent.rte")
								Missile.PresetName = "Rocket"
								Missile.Sharpness = -1
							elseif self.AimRocket.dist < 120 then
								Missile = CreateAEmitter("Mine Launch", "Independent.rte")
								Missile.PresetName = "Mine Missile"
								Missile.Sharpness = -1
							end
						elseif self.AimRocket.dist < 500 and self.AimRocket.minStr < 15 then
							Missile = CreateAEmitter("Homing Missile Launch", "Independent.rte")
							Missile.PresetName = "Homing Missile"
							Missile.Sharpness = self.AimRocket.Act.ID
						end
						
						if Missile then
							self.WatchdogTimer:Reset()
							
							Missile.Vel = Vector(5, 0):RadRotate(self.AimRocket.ang)
							Missile.RotAngle = self.AimRocket.ang
							Missile.Pos = self.Head.Pos + self.Head:RotateOffset(Vector(14,0))
							MovableMan:AddParticle(Missile)
							Missile:SetWhichMOToNotHit(self, -1)
							self:SetWhichMOToNotHit(Missile, -1)
							
							self.HeadAngle = self.AimRocket.ang
							self.missilePause = RangeRand(4500,7000)
						end
					else
						self.missilePause = RangeRand(1000,2000)
					end
					
					self.AimRocket = nil
				end
			end
		elseif self.targetIndex <= #self.Targets then
			local Targ = self.Targets[self.targetIndex]
			if Targ and Targ.dist > 175 and Targ.dist < 1500 and MovableMan:ValidMO(Targ.Act) and 
					Targ.Act.PresetName ~= "Entity" and	Targ.Act.Vel.Magnitude < 30
			then
				self.AimRocket = {Co=coroutine.create(Independent.FindRocketPath), minStr=math.huge, Act=Targ.Act, dist=0}
			end
			
			self.targetIndex = self.targetIndex + 1
		else
			self.Targets = nil
		end
	end
	
	-- Find a location to shoot from
	if self.Target and not self.FindPathCo then
		if not self.LZ or self.MoveTimer:IsPastSimMS(2000) then
			if not self.FindLZCo then
				self.FindLZCo = coroutine.create(Independent.FindLZ)
			else
				local running, LZ = coroutine.resume(self.FindLZCo, self)
				if running then
					if LZ then
						self.MoveTimer:Reset()
						self.FindLZCo = nil
						self.LZ = LZ
					end
				else
					--print("Can't find a location to shoot from")
					self.FindLZCo = nil
					self.InvTarget = self.Target
					if self.Targets and #self.Targets > 1 then
						local r = math.random(2, #self.Targets)
						if self.Targets[r] then
							self.Target = self.Targets[r].Act
							if self.Target and (not self.Target.ID or self.Target.ID > 254 or self.Target.PresetName == "Entity") then
								self.Target = nil
							end
						end
					end
				end
			end
		end
	else
		self.FindLZCo = nil
	end
	
	if self.LZ and not self.AimRocket then	-- Don't jump while evaluating targets
		if self.Vel.Magnitude < 3 then
			if SceneMan:ShortestDistance(self.Pos, self.LZ, false).Magnitude < 30 then	-- At target
				self.WatchdogTimer:Reset()
				self.LZ = nil
				self.AIMode = 0
			elseif math.abs(self.RotAngle) < 0.2 then
				if self.FindPathCo then
					if not coroutine.resume(self.FindPathCo, self) then		-- The coroutine return false when done
						self.FindPathCo = nil
						if #self.Waypoints > 0 then
							self.acc_ang = self.Waypoints[1].ang
							self.acc_t = self.Waypoints[1].acc_t
							self.TargetVel = self.Vel + Vector(self.jetBurst,0):RadRotate(self.acc_ang)
							self.Vel = self.TargetVel
							
							if self.Thrust then
								self.Thrust.ToDelete = true
							end
							
							-- Start jet pack
							self.Thrust = CreateAEmitter("Jump Thruster", "Independent.rte")
							self.Thrust.Pos = ThrusterPos
							self.Thrust.RotAngle = self.RotAngle +
								SceneMan:ShortestDistance(TurretPos + Vector(0,-15), ThrusterPos, false).AbsRadAngle
							MovableMan:AddParticle(self.Thrust)
						else
							--print("Independent: Can't find path to LZ")
							self.LZ = nil
							self.InvTarget = self.Target
							if self.Targets and #self.Targets > 1 then
								local r = math.random(2, #self.Targets)
								if self.Targets[r] then
									self.Target = self.Targets[r].Act
									if self.Target and (not self.Target.ID or self.Target.ID > 254 or self.Target.PresetName == "Entity") then
										self.Target = nil
									end
								end
							end
						end
					end
				else
					self.Waypoints = {}
					self.FindPathCo = coroutine.create(Independent.FindJumpPath)
				end
			end
		else
			self.FindPathCo = nil
		end
	end
	
	if not self.Break and self.Vel.Y > 6 and
		SceneMan:CastObstacleRay(self.Pos, self.Vel*8, Vector(), Vector(), self.ID, -1, 4) > -1
	then -- Start jet pack
		self.Break = CreateAEmitter("Break Thruster", "Independent.rte")
		self.Break.RotAngle = self.Vel.AbsRadAngle
		self.Break.Pos = self.Pos
		MovableMan:AddParticle(self.Break)
	end
	
	if self.Target then
		--DrawDebug("Eye Red", self.Target.Pos) --Debug
	
		-- Face the enemy
		Ctrl:SetState(Controller.AIM_SHARP, true)
		self:SetAimAngle(SceneMan:ShortestDistance(self.Pos, self.Target.Pos, false).AbsRadAngle)
		
		if not self.fire then
			local TargPos			
			self.even = not self.even
			if self.even then
				TargPos = self.Target.Pos + Vector(15,0):RadRotate(self.Target.RotAngle+1.571)
			else
				TargPos = self.Target.Pos
			end
			
			local Free = Vector()
			local Dist = SceneMan:ShortestDistance(TurretPos, TargPos, false)
			
			if not self.burst then
				if Dist.Magnitude < self.WepRng/3 then
					self.burst = math.random(10,20)
				else
					self.burst = math.random(4,10)
				end
			end
			
			if Dist.Magnitude < self.WepRng * 1.05 then
				local ang = Dist.AbsRadAngle
				local TestPos = TurretPos + Vector(3,0):RadRotate(ang)
				if SceneMan:GetMOIDPixel(TestPos.X, TestPos.Y) ~= self.ID then
					local id = SceneMan:CastMORay(TurretPos, Dist, self.ID, -1, false, 3)
					if id and id < 255 then
						local MO = MovableMan:GetMOFromID(MovableMan:GetRootMOID(id))
						if MO.Team ~= self.Team then
							self.WatchdogTimer:Reset()
							
							self.Flash = CreateAEmitter("Flash Blaster MG", "Independent.rte")
							self.Flash.RotAngle = ang
							self.Flash.Pos = TurretPos
							MovableMan:AddParticle(self.Flash)
							
							local Bullet = CreateMOPixel("Particle Blaster MG", "Independent.rte")
							Bullet.Pos = TestPos
							if self.Vel.Magnitude > 3 then
								Bullet.Vel = Vector(99,0):RadRotate(ang + RangeRand(-0.12, 0.12))
							else
								Bullet.Vel = Vector(99,0):RadRotate(ang + RangeRand(-0.07, 0.07))
							end
							
							Bullet:SetWhichMOToNotHit(self, -1)
							MovableMan:AddParticle(Bullet)
							
							self.burst = self.burst - 1
							if self.burst < 1 then
								self.burst = nil
								self.fire = math.random(10,30)
							else
								self.fire = math.random(5,8)
							end
						end
					end
				end
			else
				self.fire = 15
			end
		else
			self.fire = self.fire - 1
			if self.fire < 1 then
				self.fire = false
			end
		end
	end
end











-- Calculate how to jump to the LZ
Independent.FindJumpPath = function(self)
	local NextWpts = {}
	local FindJumpAngCo = coroutine.create(Independent.FindJumpAngle)
	local yi = false
	while coroutine.resume(FindJumpAngCo, self, self.Pos, NextWpts) do
		if yi then
			coroutine.yield()	-- Spread out the search over several frames
		end
		yi = not yi
	end
	
	if #NextWpts > 0 then	
		table.sort(NextWpts, function(a,b) return a.dist < b.dist end)
		if SceneMan:ShortestDistance(self.LZ, NextWpts[1].Pos, false).Magnitude < 40 then
			self.Waypoints[1] = {
				Pos = NextWpts[1].Pos,
				dist = NextWpts[1].dist,
				ang = NextWpts[1].ang,
				acc_t = NextWpts[1].acc_t
			}
			return
		else
			self.Waypoints[#self.Waypoints+1] = {
				Pos = NextWpts[1].Pos,
				dist = NextWpts[1].dist,
				ang = NextWpts[1].ang,
				acc_t = NextWpts[1].acc_t
			}
		end
		
		self.resolution = 1.5
		
		local TempWpts
		for i,Wpt in pairs(NextWpts) do	-- Now check the paths to the LZ from all positions found
			TempWpts = {}
			
			--DrawDebug("Brain", Wpt.Pos + Vector(0,-5))
			
			yi = false
			FindJumpAngCo = coroutine.create(Independent.FindJumpAngle)
			while coroutine.resume(FindJumpAngCo, self, Vector(Wpt.Pos.X, Wpt.Pos.Y), TempWpts) do
				if yi then
					coroutine.yield()	-- Spread out the search over several frames
				end
				yi = not yi
			end
			
			if #TempWpts > 0 then
				table.sort(TempWpts, function(a,b) return a.dist < b.dist end)
				
				-- Check if Wpt lead to TempWpts[1]
				if SceneMan:ShortestDistance(self.LZ, TempWpts[1].Pos, false).Magnitude < 40 then
					self.Waypoints[1] = {
						Pos = Wpt.Pos,
						dist = Wpt.dist,
						ang = Wpt.ang,
						acc_t = Wpt.acc_t
					}
					return
				else
					self.Waypoints[#self.Waypoints+1] = {
						Pos = Wpt.Pos,
						dist = TempWpts[1].dist,
						ang = Wpt.ang,
						acc_t = Wpt.acc_t
					}
				end
			end
		end
	end
	
	if #self.Waypoints > 0 then
		table.sort(self.Waypoints, function(a,b) return a.dist < b.dist end)
	elseif #NextWpts > 0 then	-- Just in case
		table.sort(NextWpts, function(a,b) return a.dist < b.dist end)
		self.Waypoints[1] = {
			Pos = NextWpts[1].Pos,
			dist = NextWpts[1].dist,
			ang = NextWpts[1].ang,
			acc_t = NextWpts[1].acc_t
		}
	end
end


-- Calculate one jump
Independent.FindJumpAngle = function(self, StartPos, Wpts)
	local Vel, Acc, NewPos, OldPos
	local Range = SceneMan:ShortestDistance(self.LZ, StartPos, false)
	local mag = Range.Magnitude
	local maxAcc = math.floor(500/TimerMan.DeltaTimeMS + 0.5) -- TTL: 500ms
	local minAcc = math.floor(math.min(1 + math.abs(Range.Y/60) + mag/60, maxAcc/2) + 0.5)
	local accRes = math.max(math.floor(5 * self.resolution + 0.5), 1)
	local minAng = 0.1
	local maxAng = 2.0
	local angRes = 0.11 * self.resolution
	
	local dir
	if Range.AbsRadAngle < 1.571 then
		dir = 3.142
		
		if self.LZ.Y < StartPos.Y then
			minAng = 0.6
			angRes = angRes * 0.9
		end
	else	-- Target is to the right
		dir = -4.712
		
		if self.LZ.Y < StartPos.Y then
			maxAng = 1.3
			angRes = angRes * 0.9
		end
	end
	
	for acc_t = minAcc, maxAcc, accRes do
		for ang = minAng, maxAng, angRes do
			Vel = self.Vel + Vector(self.jetBurst,0):RadRotate(dir-ang)
			Acc = Vector(self.jetPower,0):RadRotate(dir-ang)
			OldPos = Vector(StartPos.X, StartPos.Y)
			for t = 1, 500 do
				NewPos = OldPos + Vel * self.Factor
				--if t % 8 == 0 then DrawDebug("Blue", NewPos) end	-- Debug
				
				if t < acc_t then
					Vel = Vel + Acc + self.Gravity
				else
					Vel = Vel + self.Gravity
				end
				
				if NewPos.Y < -50 or
					SceneMan:GetTerrMatter(NewPos.X - 24, NewPos.Y - 13) > 0 or 
					SceneMan:GetTerrMatter(NewPos.X + 24, NewPos.Y - 13) > 0 or
					SceneMan:GetTerrMatter(NewPos.X, NewPos.Y + 15) > 0
				then
					break
				end
				
				-- The feet hit the ground
				if SceneMan:GetTerrMatter(NewPos.X, NewPos.Y + 20) > 0 then				
					if self.Occupied[Hash(NewPos)] then
						break
					end
					
					-- Make sure we land at a steep angle or have space to recover from a fall
					Range = SceneMan:ShortestDistance(OldPos, NewPos, false)
					if math.abs(Range.X*1.2) > math.abs(Range.Y) then
						if (Range.X > 0 and SceneMan:GetTerrMatter(NewPos.X + 30, NewPos.Y + 20) == 0) or
							(Range.X < 0 and SceneMan:GetTerrMatter(NewPos.X - 30, NewPos.Y + 20) == 0)
						then
							break
						end
					end
					
					Wpts[#Wpts+1] = {
						Pos = Vector(NewPos.X, NewPos.Y),
						dist = SceneMan:ShortestDistance(self.LZ, NewPos, false).Magnitude + t/10,
						ang = dir - ang,
						acc_t = acc_t
					}
					
					--DrawDebug("Cross", Vector(NewPos.X, NewPos.Y))
					
					break
				end
				
				OldPos = Vector(NewPos.X, NewPos.Y)
			end
			
			coroutine.yield(true)	-- Spread out the search over several frames
		end
	end
end

-- Try to find a rocket trajectory to the target
Independent.FindRocketPath = function(self, TargetPos)
	local OldPos, NewPos, Vel
	local dist, angle, state, hs
	
	local StartPos = self.Pos + self:RotateOffset(Vector(7,-20))
	local fiveHundredMS = 500 / TimerMan.DeltaTimeMS
	local delta = FrameMan.PPM * TimerMan.DeltaTimeSecs * 3	-- Simulate every third frame
	
	local Temp
	local tmp = SceneMan:ShortestDistance(StartPos, TargetPos, false).Magnitude
	if StartPos.Y > TargetPos.Y then	-- The target is above the shooter
		Temp = Vector(TargetPos.X, TargetPos.Y - tmp)
	else
		Temp = Vector(TargetPos.X, TargetPos.Y - tmp/3)
	end
		
	local startAng = SceneMan:ShortestDistance(StartPos, Temp, false).AbsRadAngle
	for aimAng = startAng-0.5, startAng+0.5, 0.08 do
		angle = aimAng
		state = 1	-- Not armed
		Vel = Vector(5, 0):RadRotate(angle)
		OldPos = Vector(StartPos.X, StartPos.Y)
		
		for t = 1, 93, 3 do
			NewPos = OldPos + Vel * delta
			
			hs = Hash(NewPos)
			if self.Occupied[hs] and self.Occupied[hs] < 0 then	-- Avoid hitting allies
				break
			end
			
			if t <= fiveHundredMS then
				Vel = Vel + Vector(0.8,0):RadRotate(angle) * 3	-- Fake rocket emitter
			else
				--DrawDebug("Orange", NewPos)	-- Debug
				
				state = 2 -- Armed
				Temp = SceneMan:ShortestDistance(NewPos, TargetPos, false)
				dist = Temp.Magnitude
				
				-- Taget bracketed?
				if math.abs(SceneMan:ShortestDistance(OldPos, TargetPos, false).AbsRadAngle - Temp.AbsRadAngle) > 1 then
					Temp = OldPos + SceneMan:ShortestDistance(OldPos, NewPos, false) * 0.5
					dist = SceneMan:ShortestDistance(Temp, TargetPos, false).Magnitude
					if dist < 30 then
						state = 3 -- This will probably hit the target
						break
					end
				end
			end
			
			-- The missile hit something
			if NewPos.Y < -25 or SceneMan:GetTerrMatter(NewPos.X, NewPos.Y) > 0 then
				break
			end
			
			for _ = 1, 3 do	-- Fake OrientToVel
				angle = angle*0.67 + Vel.AbsRadAngle*0.33
			end
			
			Vel = Vel + Vector(0, 0.66) -- Fake gravity	(0.2*3.3)
			OldPos:SetXY(NewPos.X, NewPos.Y)
		end
		
		if state > 2 then
			dist = dist * 0.5
			tmp = 0
		elseif state > 1 then
			if dist < 500 then
				tmp = SceneMan:CastStrengthSumRay(NewPos, TargetPos, 4, 128)	-- Ignore grass
			else
				tmp = 10000
			end
		else
			dist = 10000
			tmp = 10000
		end
		
		coroutine.yield({ang=aimAng, dist=dist, str=tmp})	-- Spread out the search over several frames
	end
end

-- Find a location to shoot from
Independent.FindLZ = function(self)
	local Free = Vector()
	local TestPos, dist
	local LZs = {}
	local startAng = SceneMan:ShortestDistance(self.Target.Pos, self.Pos, false).AbsRadAngle * 0.67
	for ang = startAng, startAng + 6.28, 0.13 do
		for rng = 9, self.WepRng, 9 do
			TestPos = self.Target.Pos + Vector(rng,0):RadRotate(ang)
			--DrawDebug("Orange", TestPos)
			if TestPos.Y < 0 or SceneMan:GetTerrMatter(TestPos.X, TestPos.Y) > 0 then
				break
			end
			
			if rng > 55 then
				if SceneMan:CastObstacleRay(TestPos, Vector(0,30), Vector(), Free, self.ID, -1, 5) > 0 and
					not self.Occupied[Hash(Free)]
				then
					dist = SceneMan:ShortestDistance(self.Pos, Free, false).Magnitude / 2
					dist = dist - SceneMan:ShortestDistance(self.Target.Pos, Free, false).Magnitude
					dist = dist + SceneMan:CastStrengthSumRay(Free, Free + Vector(0,-60), 3, 128)
					table.insert(LZs, {Pos=Vector(Free.X, Free.Y - 20), score=dist})
				end
			end
		end
		
		coroutine.yield()	-- Spread out the search over several frames
	end
	
	if #LZs > 0 then
		table.sort(LZs, function(a,b) return a.score < b.score end)
		coroutine.yield(LZs[1].Pos) -- Return the answer
	end
end